/*-
 * Copyright (c) 2006 Allan Saddi <allan@saddi.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: main.c 2242 2006-12-13 21:32:35Z asaddi $
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <libgen.h>
#include <signal.h>
#include <fcntl.h>
#include <errno.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <pthread.h>

#include "ajp.h"
#include "wsgi.h"
#include "version.h"

static char *progName;
static int verbosity = 0;

struct clientData {
  int sock;
  struct sockaddr_in addr;
};

static void *
clientHandler(void *clientData)
{
  struct clientData *cd = clientData;
  AJPContext *ctxt;

  if (verbosity)
    fprintf(stderr, "Connection opened (%s:%d)\n",
	    inet_ntoa(cd->addr.sin_addr),
	    (int)ntohs(cd->addr.sin_port));
  if ((ctxt = ajpCreateContext(cd->sock, wsgiHandler, wsgiBodyHandler))) {
    /* Main loop */
    while (ajpProcessInput(ctxt) == AJP_OK);

    ajpDestroyContext(ctxt);
  }
  if (verbosity)
    fprintf(stderr, "Connection closed (%s:%d)\n",
	    inet_ntoa(cd->addr.sin_addr),
	    (int)ntohs(cd->addr.sin_port));

  free(cd);
  return NULL;
}

static int
createSocket(int port)
{
  int sock = -1;
  struct sockaddr_in addr;
  int on = 1;

  if ((sock = socket(AF_INET, SOCK_STREAM, PF_UNSPEC)) == -1)
    goto bad;

  if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on)))
    goto bad;

  memset(&addr, 0, sizeof(addr));
  addr.sin_len = sizeof(addr);
  addr.sin_family = AF_INET;
  addr.sin_port = htons(port);
  addr.sin_addr.s_addr = htonl(INADDR_LOOPBACK);

  if (bind(sock, (struct sockaddr *)&addr, sizeof(addr)))
    goto bad;

  if (listen(sock, SOMAXCONN))
    goto bad;

  return sock;

 bad:
  if (sock != -1)
    close(sock);
  return -1;
}

static void
die(int foo)
{
  exit(0);
}

static void
reload(int foo)
{
  exit(42);
}

static void
usage(void)
{
  fprintf(stderr, "Usage: %s [-BVv] [-l <logFile>] [-p <port>] <moduleName> <appName> [<scriptName>]\n",
	  progName);
  exit(1);
}

int
main(int argc, char *argv[])
{
  int ch;
  char *end;
  char *scriptName = "";
  int port = 8009;
  int daemonize = 0;
  char *logName = NULL;

  int sock;

  progName = strdup(basename(argv[0]));

  while ((ch = getopt(argc, argv, "BVl:p:v")) != -1) {
    switch (ch) {
    case 'B':
      daemonize++;
      break;
    case 'V':
      fprintf (stderr, "%s %s\n", progName, VERSION_STRING);
      exit(1);
      break;
    case 'l':
      logName = optarg;
      break;
    case 'p':
      port = strtol(optarg, &end, 10);
      if (*end) {
	fprintf (stderr, "%s: bad port argument\n", progName);
	exit(1);
      }
      break;
    case 'v':
      verbosity++;
      break;
    default:
      usage();
    }
  }
  argc -= optind;
  argv += optind;

  if (argc < 2 || argc > 3)
    usage();

  if (argc > 2) {
    scriptName = argv[2];
    if (scriptName[0]) {
      if (scriptName[0] != '/') {
	fprintf(stderr, "%s: scriptName must start with a '/'\n", progName);
	exit(1);
      }
      if (scriptName[strlen(scriptName) - 1] == '/') {
	fprintf(stderr, "%s: scriptName must not end with a '/'\n", progName);
	exit(1);
      }
    }
  }

  if ((sock = createSocket(port)) == -1) {
    perror(progName);
    exit(2);
  }

  if (logName != NULL) {
    int log;

    if ((log = open(logName, O_CREAT|O_APPEND|O_WRONLY, 0666)) == -1) {
      fprintf(stderr, "%s: %s: %s\n", progName, logName, strerror(errno));
      exit(3);
    }

    dup2(log, STDERR_FILENO);

    if (log != STDERR_FILENO)
      close(log);
  }

  if (daemonize) {
    /* Do this ourselves so we have more control */
    int fd = open("/dev/null", O_RDWR);
    dup2(fd, STDIN_FILENO);
    dup2(fd, STDOUT_FILENO);
    if (logName == NULL)
      dup2(fd, STDERR_FILENO);
    if (fd > STDERR_FILENO)
      close(fd);
  }

  if (wsgiInit(argv[0], argv[1], scriptName, progName))
    exit(4);

  if (daemonize) {
    if (daemon (1, 1)) {
      perror(progName);
      exit(5);
    }
  }
  else
    signal(SIGHUP, reload);

  signal(SIGINT, die);
  signal(SIGTERM, die);

  for(;;) {
    struct clientData *cd = malloc(sizeof(*cd));
    socklen_t addrLen = sizeof(cd->addr);

    if ((cd->sock = accept(sock, (struct sockaddr *)&cd->addr,
			   &addrLen)) != -1) {
      pthread_t t;
      if (pthread_create(&t, NULL, clientHandler, cd)) {
	close(cd->sock);
	free(cd);
      }
      else
	pthread_detach(t);
    }
  }
}
